import os
import shutil

"""
Plugin project generator
"""

here = os.path.dirname(__file__)
demo_root = os.path.abspath(os.path.join(here, 'demo_project'))


class PluginProjectBuilder:

    def build(self, path=None):
        if not path:
            path = os.path.abspath('./demo')
        if os.path.exists(path):
            for demo_item in os.listdir(demo_root):
                src = os.path.abspath(os.path.join(demo_root, demo_item))
                dst = os.path.abspath(os.path.join(path, demo_item))
                if os.path.isdir(src):
                    shutil.copytree(src, dst)
                else:
                    shutil.copy2(src, dst)
        else:
            shutil.copytree(demo_root, path)



import base64
import gzip
import json

from lyrebird.mock import context

class CacheHandler:
    """
    将当前request和response保存至application.cache中
    内存中保存的请求格式如下:
    {
        "request": {
            "data": null,
            "headers": {
                "Accept": "*/*",
                "Accept-Encoding": "gzip, deflate",
                "Connection": "keep-alive",
                "Host": "localhost:9090",
                "User-Agent": "python-requests/2.12.4"
            },
            "method": "GET",
            "url": "http://localhost:9090/mock/http://www.baidu.com"
        },
        "response": {
            "code": 200,
            "data": "<html>...</html>",
            "headers": {
                "Cache-Control": "private, no-cache, no-store, proxy-revalidate, no-transform",
                "Content-Type": "text/html",
                "Date": "Tue, 23 May 2017 07:03:59 GMT",
                "Last-Modified": "Mon, 23 Jan 2017 13:27:31 GMT",
                "Pragma": "no-cache",
                "Server": "bfe/1.0.8.18",
                "Set-Cookie": "BDORZ=27315; max-age=86400; domain=.baidu.com; path=/"
        }
    }
    """
    def handle(self, handler_context):
        handler_context.update_client_resp_time()
        flow = self.ctx_to_dict(handler_context)
        context.application.cache.add(flow)
        context.application.socket_io.emit('action', 'add flow log')
        if context.application.work_mode == context.Mode.RECORD:
            mock_tag = flow['response']['headers']['lyrebird']
            if mock_tag.startswith('proxy'):
                group = context.application.data_manager.current_data_group
                if group:
                    group.add_data_and_filter(flow)


    def ctx_to_dict(self, handler_context):
        ctx = dict()
        ctx['id'] = handler_context.id
        ctx['time'] = handler_context.client_req_time
        ctx['response-time'] = handler_context.client_resp_time - handler_context.client_req_time
        if handler_context.request:
            req = dict()

            url = handler_context.get_origin_url()
            if url:
                req['url'] = url
            else:
                req['url'] = handler_context.request.url

            req['headers'] = {k: v for k, v in handler_context.request.headers.items()}

            # data 处理
            # 解压gzip内容
            content_type = handler_context.request.headers.get('Content-Type')
            user_agent = handler_context.request.headers.get("User-Agent")
            if 'Content-Encoding' in handler_context.request.headers \
                    and handler_context.request.headers['Content-Encoding'] == 'gzip':
                data = gzip.decompress(handler_context.request.data)
            elif handler_context.request.data[:3] == b'\x1f\x8b\x08':
                data = gzip.decompress(handler_context.request.data)
            else:
                data = handler_context.request.data
            # 保存data
            req['data'] = self.data_2_str(content_type, user_agent, data)

            req['method'] = handler_context.request.method
            ctx['request'] = req

        if handler_context.response:
            resp = dict()
            resp['code'] = handler_context.response.status_code
            resp['headers'] = {k: v for k, v in handler_context.response.headers.items()}
            content_type = handler_context.response.headers.get('Content-Type')
            user_agent = handler_context.request.headers.get("User-Agent")
            resp['data'] = self.data_2_str(content_type, user_agent, handler_context.response.data)
            ctx['response'] = resp
        return ctx

    def data_2_str(self, content_type, user_agent, data):
        if not content_type or len(data) == 0:
            return
        try:
            if 'json' in content_type:
                return json.loads(data.decode())
            elif 'text' in content_type or 'xml' in content_type:
                charset_index = content_type.find('charset')
                if charset_index >= 0:
                    charset = content_type[charset_index+8:].strip()
                    return data.decode(charset)
                else:
                    return data.decode()
            elif 'form' in content_type:
                return data.decode()
            else:
                try:
                    return json.loads(data.decode())
                except Exception:
                    return data.decode()
        except Exception:
            return 'Unknown byte data'



import json
import codecs
import uuid
from pathlib import Path
import shutil
from .data import Data
from collections import OrderedDict
from .exceptions import DataGroupNotExistsError, DataGroupIsNotDirError, DataGroupInfoNotFoundError
from lyrebird.log import get_logger


logger = get_logger()


class Group:
    """
    DataGroup

    The info file (.lyrebird_prop) contains group info.
    {
        "id": "",
        "name": ""
    }

    """

    info_filename = '.lyrebird_prop'


    def __init__(self, gid, gname, gpath, gparent_id=None):
        self.id = gid
        self.name = gname
        self.parent_id = gparent_id
        self.path = gpath
        self.all_data = OrderedDict()

    @classmethod
    def createify(cls, group_path):
        """
        Create group object if group_path exists
        """
        new_path = Path(group_path)
        if not new_path.exists():
            raise DataGroupNotExistsError(group_path)
        if not new_path.is_dir():
            raise DataGroupIsNotDirError(group_path)
        # Read group info
        with codecs.open(new_path/cls.info_filename, 'r' , 'utf-8') as f:
            info = json.load(f)
        if not info:
            raise DataGroupInfoNotFoundError(new_path)

        _gid = info.get('id')
        _gname = info.get('name')

        group = cls(_gid, _gname, new_path)
        group.scan()
        return group

    @classmethod
    def new_group(cls, group_dir):
        """
        Make a new group
        """
        _id = str(uuid.uuid4())
        group = cls(_id, _id, Path(group_dir)/_id)
        Path(group.path).mkdir(parents=True, exist_ok=True)
        return group

    def save(self):
        with codecs.open(self.path/self.info_filename, 'w', 'utf-8') as f:
            info = {
                'id': self.id,
                'name': self.name,
                'parent': self.parent_id
            }
            json.dump(info, f, ensure_ascii=False, indent=4)
    
    def delete(self):
        """
        Delete group dir and subfile from disk
        """
        shutil.rmtree(self.path)

    def create_data(self, flow=None):
        data = Data.new_data(self.path, flow)
        self.all_data[data.id] = data
        return data
    
    def delete_data(self, data_id):
        data = self.all_data.pop(data_id)
        data.delete()

    def scan(self):
        for subfile in Path(self.path).iterdir():
            if not subfile.is_dir():
                continue
            try:
                data = Data.createify(subfile)
                if data:
                    self.all_data[data.id] = data
            except Exception as e:
                logger.error(f'Load data failed. {subfile}\n{e}')    

    def json(self, detail=False):
        json_obj = {
            'id': self.id,
            'name': self.name,
            'parent': self.parent_id,
        }
        if detail:
            json_obj['data_list'] = [self.all_data[data_id].json() for data_id in self.all_data]
        return json_obj



from flask_restful import Resource
from lyrebird.mock import context, plugin_manager
from flask import request, jsonify, abort
from lyrebird import application


class Conf(Resource):
    """
    Lyrebird 及 插件 配置文件获取和修改
    """

    def get(self, plugin_name):
        if plugin_name == 'lyrebird':
            return jsonify(application.config.raw())
        else:
            return jsonify(plugin_manager.get_conf(plugin_name))

    def put(self, plugin_name):
        try:
            if plugin_name == 'lyrebird':
                context.application.conf = request.get_json()
                context.application.save()
                return context.make_ok_response()
            else:
                plugin_manager.set_conf(plugin_name, request.get_json())
                return context.make_ok_response()
        except Exception as e:
            return context.make_fail_response(str(e))


class ResetConf(Resource):
    """
    Lyrebird 及 插件 配置文件重置
    """
    def put(self, plugin_name):
        try:
            if plugin_name == 'lyrebird':
                return context.make_fail_response('暂未开放')
            else:
                plugin_manager.set_default_conf(plugin_name)
                return context.make_ok_response()
        except Exception as e:
            return context.make_fail_response(str(e))



from lyrebird.mock import context
from flask import Response, stream_with_context
import json


class MockHandler:
    """
    根据当前设置数据组的匹配条件,查找对应的mock数据。
    如果没有找到匹配的数据则交由下一个处理器处理。

    """
    def handle(self, handler_context):
        data = context.application.data_manager.router.get_mock_data(handler_context.flow)
        if data:
            handler_context.response = self.data2response(data)

    def data2response(self, data):
        resp_info = json.loads(data.response.content)
        code = resp_info['code']
        headers = resp_info['headers']
        headers['lyrebird'] = 'mock'
        resp_data = data.response_data.content
        def gen():
            yield resp_data
        return Response(stream_with_context(gen()), status=code, headers=headers)



from jinja2 import Environment, PackageLoader
from flask import send_from_directory
import sys
import os
import inspect
import codecs


class Rule:
    """
    视图路由规则

    """

    def __init__(self, rule, endpotion, view_func=None, **options):
        self.rule = rule
        self.endpoint = endpotion
        self.view_func = view_func
        self.options = options

    def __str__(self):
        return f'Rule {self.rule} func={self.view_func}'


class EventRule:
    """
    socket-io事件规则
    """

    def __init__(self, event, handler, namespace):
        self.event = event
        self.handler = handler
        self.namespace = namespace

    def __str__(self):
        return f'Event rule {self.event} handler={self.handler}'


class PluginView:
    """
    插件视图基类

    """

    def __init__(self):
        self.env = Environment()
        self.rules = []
        self.event_rules = []
        self.static_folder = None
        self.template_folder = None

    def add_url_rule(self, rule, endpoint=None, view_func=None, **options):
        self.rules.append(Rule(rule, endpoint, view_func, **options))

    def on_event(self, event, handler, namespace):
        self.event_rules.append(EventRule(event, handler, namespace))

    def on_create(self):
        """
        初始生命周期
        通过在该函数内调用'set_template_root'指定前端模板目录
        通过在该函数内调用'set_static_root'指定前端静态文件目录
        通过在该函数内调用'add_url_rule'添加处理请求的路由，插件的主页面要使用'/'路径
        """
        pass

    def after_on_create(self):
        if not self.static_folder:
            self.set_static_root()
        if not self.template_folder:
            self.set_template_root()

    def _get_current_package_name(self):
        """
        通过class判断包名
        :return: 
        """
        module_name = self.__class__.__module__
        if '.' in module_name:
            return module_name.split('.')[0]
        return module_name

    def get_package_file_path(self, file_path, package_name=None):
        if not package_name:
            package_name = self._get_current_package_name()
        target_module = sys.modules.get(package_name)
        package_path = os.path.dirname(inspect.getsourcefile(target_module))
        return os.path.abspath(os.path.join(package_path, file_path))

    def set_template_root(self, package_name=None, templates_path='templates'):
        """
        设置html模板根目录. 默认在 [插件包]/templates 下

        :param package_name: 包名，不指定时默认读取当前插件包 
        :param templates_path: 模板路径， 默认为templates目录
        """
        if not package_name:
            package_name = self._get_current_package_name()
        self.env.loader = PackageLoader(package_name, package_path=templates_path)
        target_module = sys.modules.get(package_name)
        package_path = os.path.dirname(target_module.__file__)
        self.template_folder = os.path.abspath(os.path.join(package_path, templates_path))

    def set_static_root(self, package_name=None, static_path='static'):
        """
        设置静态文件(css、js等)目录. 默认在 [插件包]/static 下

        :param package_name: 包名，不指定时默认读取当前插件包 
        :param static_path: 静态文件目录， 默认为static目录
        """
        if not package_name:
            package_name = self._get_current_package_name()
        target_module = sys.modules.get(package_name)
        package_path = os.path.dirname(target_module.__file__)
        self.static_folder = os.path.abspath(os.path.join(package_path, static_path))
        self.add_url_rule('/static/<path:filename>', view_func=self.send_static_file)

    def render_template(self, template, *args, **kwargs):
        """
        渲染模板，支持jinja2语法

        :param template: 模板文件名 
        :param args: 参数数组
        :param kwargs: 参数字典
        :return: 返回渲染完成的html文本
        """
        return self.env.get_template(template).render(*args, **kwargs)

    def send_static_file(self, filename):
        """
        响应静态文件
        收到请求后，按照filename在静态文件目录下查找，并返回静态文件

        :param filename: 
        :return: 
        """
        # todo cache_timeout
        return send_from_directory(self.static_folder, filename)

    def send_template_file(self, filename):
        template_file_path = os.path.abspath(os.path.join(self.template_folder, filename))
        return codecs.open(template_file_path, 'r', 'utf-8').read()

    def default_conf(self):
        """
        设置默认的 conf.json

        :return: 返回 conf.json 内容
        """
        # todo plugin conf 应该以对象方式读取存储 并且对外只读，特定方法支持写
        return {}



from flask_restful import Resource
from flask import jsonify
from lyrebird.mock import context
from lyrebird.mock import plugin_manager


class Menu(Resource):

    def get(self):
        menu = [
            {
                'name': 'inspector',
                'title': 'Inspector',
                'type': 'router',
                'path': '/'
            },
            {
                'name': 'datamanager',
                'title': 'DataManager',
                'type': 'router',
                'path': '/datamanager'
            }]
        for plugin_key in plugin_manager.plugins:
            plugin = plugin_manager.plugins[plugin_key]
            if 'beta_web' not in plugin:
                continue
            web = plugin['beta_web']
            name = plugin['name']
            if hasattr(web, 'get_title'):
                title = web.get_title()
            else:
                title = name[0].upper() + name[1:]
            menu.append({
                'name': 'plugin-view',
                'title': title,
                'type': 'router',
                'path': '/plugin',
                'params': {
                    'src': f'/plugin/{plugin["project_name"]}',
                    'name': name
                }
            })
        return context.make_ok_response(menu=menu)



import json
import codecs
from pathlib import Path
from bs4 import BeautifulSoup
from xml.dom import minidom


class DataFile:
    
    def __init__(self, data_obj, name, filetype='bin'):
        self.name = name
        self.path = Path(data_obj.path)/name
        self.filetype = filetype

    def load(self):
        _content = None
        if self.filetype in ['json', 'text', 'html', 'xml']:
            with codecs.open(self.path, 'r', 'utf-8') as f:
                _content = f.read()
        else:
            with codecs.open(self.path, 'rb', 'utf-8') as f:
                _content = f.read()
        return _content
    
    def save(self, val):
        if self.filetype in ['json', 'html', 'text', 'xml']:
            with codecs.open(self.path, 'w', 'utf-8') as f:
                f.write(val)

    @property
    def content(self):
        try:
            return self.load()
        except Exception:
            return None

    @content.setter
    def content(self, val):
        if not val:
            return

        _content = None
        if self.filetype == 'json':
            _content = json.dumps(json.loads(val), ensure_ascii=False, indent=4)
        elif self.filetype == 'html':
            _content = BeautifulSoup(val, features="html.parser").prettify()
        elif self.filetype == 'text':
            _content = val
        elif self.filetype == 'xml':
            _content = minidom.parseString(val).toprettyxml()

        if _content:
            self.save(_content)

    def json(self):
        return {
            'content': self.content,
            'filetype': self.filetype
        }


from . import cache_handler, mock_handler, path_not_found_handler, proxy_handler
from collections import OrderedDict


def get_inner_handlers():
    return OrderedDict(
        {
            'mock': mock_handler.MockHandler(),
            'proxy': proxy_handler.ProxyHandler(),
            '404': path_not_found_handler.RequestPathNotFound()
        }
    )



from flask import jsonify, stream_with_context, Response
from . import cache
from .filesystem import FileManager
from .data_manager import DataManager
from flask_socketio import SocketIO
import codecs, json, os
from lyrebird import application as app
import time


"""
Mock server context

"""

class Mode:
    NORMAL = 'normal'
    RECORD = 'record'

    @staticmethod
    def contains(val):
        if val==Mode.NORMAL or val==Mode.RECORD:
            return True
        else:
            return False 

class Application:

    def __init__(self):
        self._conf = None
        # todo 使用内存中的List存储请求，应可支持切换redis
        self.cache = cache.get_cache()
        self.work_mode = Mode.NORMAL
        # todo 使用文件系统存储mock数据，应可支持切换redis，mysql
        # self.data_manager = FileManager()
        self.data_manager = DataManager()
        # SocketIO
        self.socket_io: SocketIO = None
        self.conf_manager = None
        self.current_conf_name = None

    @property
    def event_bus(self):
        return app.server.get('event')

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, _conf):
        self._conf = _conf
        # TODO 更新conf触发更新data_manager根目录
        if _conf.get('mock.data'):
            self.data_manager.root = _conf.get('mock.data')

    def save(self):
        DEFAULT_CONF = os.path.join(
            os.path.join(os.path.expanduser('~'), '.lyrebird'), 'conf.json')
        # TODO Lyrebird 的 conf.json 应该有恢复机制或补救措施。
        with codecs.open(DEFAULT_CONF, 'w', 'utf-8') as f:
            f.write(json.dumps(self._conf, ensure_ascii=False, indent=4))


"""
SocketIO emit interval
Because of iview table has render preformance problem
We need to limit render time
"""
EMIT_INTERVAL = 0.4
last_emit_time = {}


application = Application()
db = None

def make_ok_response(**kwargs):
    ok_resp = {
            "code": 1000,
            "message": "success"
        }
    ok_resp.update(kwargs)
    return jsonify(ok_resp)


def make_fail_response(msg, code=3000, **kwargs):
    fail_resp = {
            "code": code,
            "message": msg
        }
    fail_resp.update(kwargs)
    return jsonify(fail_resp)


def make_streamed_response(generator, code=200, mimetype='application/json'):
    return Response(stream_with_context(generator()), mimetype=mimetype, status=code)


def emit(event, *args, **kwargs):
    now = time.time()
    last_push_time = last_emit_time.get(event, 0)
    if (now - last_push_time) > EMIT_INTERVAL:
        application.socket_io.emit(event, *args, **kwargs)
        last_emit_time[event] = now



from flask_restful import Resource
from lyrebird.mock import context
from flask import request, jsonify, abort, stream_with_context
import json
import time


class Flow(Resource):
    """
    当前请求单条数据
    """
    def get(self, id):
        for item in context.application.cache.items():
            if item['id'] == id:
                return jsonify(item)
        return abort(400, 'Request not found')


class FlowList(Resource):
    """
    当前请求列表
    """
    def get(self):
        all_items = context.application.cache.items()[::-1]
        req_list = []
        for item in all_items:
            info = dict(
                id=item['id'], 
                size=item['size'],
                duration=item['duration'], 
                request=dict(
                    url=item['request'].get('url'),
                    path=item['request'].get('path'),
                    host=item['request'].get('host')
                    ),
                response=dict(
                    code=item['response']['code'],
                    mock=item['response']['headers'].get('lyrebird', 'proxy')
                    )if item.get('response') else {}
                )
            req_list.append(info)
        def gen():        
            yield json.dumps(req_list)
        return context.make_streamed_response(gen)

    def delete(self):
        _ids = request.json.get('ids')
        if _ids:
            context.application.cache.delete_by_ids(*_ids)
        else:
            context.application.cache.clear()
        context.application.socket_io.emit('action', 'delete flow log')
        return context.make_ok_response()

    def post(self):
        _ids = request.json.get('ids')
        record_items = []
        for _id in _ids:
            for item in context.application.cache.items():
                if _id == item['id']:
                    record_items.append(item)
                    break
        dm = context.application.data_manager
        activated_group = dm.groups.get(dm.activated_group_id)
        if not activated_group:
            return context.make_fail_response('Not activate any group')
        
        flow_list = context.application.cache.items()
        for flow in flow_list:
            if flow['id'] in _ids:
                activated_group.create_data(flow=flow)

        return context.make_ok_response()



import os
import sys
import time
import json
import errno
import socket
import datetime
import subprocess
from . import plugin_manager
from flask import Flask, request, redirect, url_for, Response
from . import context
from .blueprints.plugin import plugin
from .blueprints.apis import api
from .blueprints.ui import ui
from .blueprints.api_mock import api_mock
from flask_socketio import SocketIO
from .reporter import report_handler
from ..version import VERSION
from lyrebird.base_server import ThreadServer
from lyrebird import application
from lyrebird import log
from lyrebird.mock.db.database import DataBase


"""
Mock server

Lyrebird main server
Default port : 9090

* HTTP mock
* HTTP record
* Lyrebird UI
* Lyrebird API
* Lyrebird plugin management
"""

_logger = log.get_logger()


class LyrebirdMockServer(ThreadServer):
    """
    模拟接口服务
    使用flask在默认的9090端口启动，模拟线上接口，同时支持通过api动态修改接口数据。

    """
    def __init__(self):
        super().__init__()
        
        self.conf = application.config
        # TODO rm conf rom mock context
        context.application.conf = application.config

        self.debug = False
        self.port = 9090
        self._working_thread = None
        self.app = Flask('MOCK')
        
        self.app.jinja_env.block_start_string = '[%'
        self.app.jinja_env.block_end_string = '%]'
        self.app.jinja_env.variable_start_string = '[['
        self.app.jinja_env.variable_end_string = ']]'
        self.app.jinja_env.comment_start_string = '[#'
        self.app.jinja_env.comment_end_string = '#]'

        # Add global function for templates
        self.app.jinja_env.globals['time'] = time.time
        self.app.jinja_env.globals['datetime'] = datetime.datetime
        self.app.jinja_env.globals['version'] = VERSION

        # async_mode = threading / eventlet / gevent / gevent_uwsgi
        self.socket_io = SocketIO(self.app, async_mode='threading', log_output=False)
        # 存储socket-io
        context.application.socket_io = self.socket_io
        # 生成过滤器实例
        if self.conf:
            self.port = self.conf.get('mock.port')
        else:
            _logger.error('Can not start mock server without config file')
            raise SyntaxError('Can not start mock server without config file.'
                              ' Default config file path = api-mock/conf.json')

        # sqlite初始化
        ROOT_DIR = application.root_dir()
        DB_FILE_NAME = 'lyrebird.db'
        if ROOT_DIR:
            SQLALCHEMY_DATABASE_URI = ROOT_DIR/DB_FILE_NAME
            # TODO: 'sqlite:///' is unfriendly to windows
            self.app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///'+str(SQLALCHEMY_DATABASE_URI)
            self.app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
            context.db = DataBase(self.app)
        
        # Plugin
        # init plugin
        plugin_manager.load()
        # load plugin frontend
        plugin_manager.add_view_to_blueprint(ui)
        plugin_manager.add_view_to_blueprint(plugin)
        # Register event socket
        plugin_manager.add_event_rules(self.socket_io)
        # Register blueprints
        self.app.register_blueprint(api)
        self.app.register_blueprint(api_mock)
        self.app.register_blueprint(ui)
        self.app.register_blueprint(plugin)
        
        @self.app.route('/')
        def index():
            """
            设置默认页面为UI首页
            """
            return redirect(url_for('ui.index'))

        @self.app.after_request
        def after_request(response: Response):
            """
            输出每条请求概要信息
            """
            lyrebird_info = response.headers.get('lyrebird', default='')
            _logger.info(f'{response.status_code} {lyrebird_info} {request.method} {request.url[:100]}')
            return response

    def run(self):
        server_ip = application.config.get('ip')
        _logger.warning(f'start on http://{server_ip}:{self.port}')
        report_handler.start()
        # cannot import at beginning, cause db hasn't init
        from lyrebird.mock.db.models import active_db_listener
        active_db_listener()
        self.socket_io.run(self.app, host='0.0.0.0', port=self.port, debug=self.debug, use_reloader=False)


    def stop(self):
        """
        停止服务

        """
        super().stop()
        try:
            self.socket_io.stop()
        except Exception:
            pass
        report_handler.stop()
        _logger.warning('MockServer shutdown')
